/*  Botik::Key Pro - Botik Key Client for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2004 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include <cmnfuncs.h>

////////////////////////////////////////////////////////////////////////
// Debug functions and definitions.
////////////////////////////////////////////////////////////////////////

#ifndef cfDONT_USE_DPRINTF
VOID cfdprintf(LPCTSTR szFormat_, ...)
{
    LPTSTR  szPreOutput, szOutput;
    va_list pArgs;
	INT     nLen;

    va_start(pArgs, szFormat_);

	nLen = _vsctprintf(szFormat_, pArgs);

	szPreOutput = new TCHAR[nLen + 1];

	_vstprintf(szPreOutput, szFormat_, pArgs);

	va_end(pArgs);

	szOutput = new TCHAR[nLen + _tcslen(cfDPRINTF_APP_NAME) + _tcslen(cfDPRINTF_OUTPUT_FORMAT) + 1];

	_stprintf(szOutput, cfDPRINTF_OUTPUT_FORMAT, cfDPRINTF_APP_NAME, szPreOutput);

    OutputDebugString(szOutput);

	delete[] szPreOutput;
	delete[] szOutput;
};
#endif // cfDONT_USE_DPRINTF

////////////////////////////////////////////////////////////////////////
// File operations...
////////////////////////////////////////////////////////////////////////
DWORD cfReadLine(HANDLE hFile_, LPTSTR* lpszBuf_, DWORD dwBufSize_, LPDWORD lpdwOutBufSize_)
{
	CONST DWORD	dwChunkSize = 128;
	DWORD		dwRetSize = 0;
	TCHAR		szChunkBuf[dwChunkSize + 1];
	DWORD		dwReaded = 0;			// in bytes
	LPTSTR		lpBuf = *lpszBuf_;
	LPTSTR		lpPtr;
	LONG		lFilePos = 0;
	DWORD		dwBufSize = dwBufSize_;	// buf size in TCHAR
	DWORD		dwChunkLen = 0; // Number of TCHARs readed from file.
	DWORD		dwDataLen = 0;  // Number of TCHARs in the buffer.

	CONST TCHAR	lf = _T('\n');
	CONST TCHAR	cr = _T('\r');

	cfdprintf(_T("cfReadLine"));

	cfdprintf(_T("cfReadLine: lpBuf: %d; BufSize: %d"), *lpszBuf_, dwBufSize_);

	if(lpBuf != NULL) 
		lpBuf[0] = _T(0);

	while(1)
	{
		if(!ReadFile(hFile_, szChunkBuf, dwChunkSize*sizeof(TCHAR), &dwReaded, NULL))
			break;

		if(dwReaded == 0)
			break;

		dwChunkLen = dwReaded/sizeof(TCHAR);

		szChunkBuf[dwChunkLen] = 0; // Set zero terminator.

		// Check if we have a enough free space for new data chunk in the buffer.
		// If it less than available data, increase it.
		if(lpBuf == NULL || dwChunkLen > (dwBufSize - dwDataLen))
		{
			DWORD dwNewSize = dwBufSize + dwChunkLen + dwChunkSize + 1;
			lpBuf = (LPTSTR)cfRealloc(lpBuf, dwBufSize*sizeof(TCHAR), dwNewSize*sizeof(TCHAR));
			dwBufSize = dwNewSize;
		};

		_tcscat(lpBuf, szChunkBuf);
		dwDataLen += dwChunkLen;

		// Find \n char in the last added data chunk...
		lpPtr = _tcschr(lpBuf + (dwDataLen - dwChunkLen), lf);

		//if(!lpPtr) // If there is no \n char, check for \r
		//	lpPtr = _tcschr(szChunkBuf, cr);

		if(lpPtr) // If there is \n char, set it to zero.
		{
			*lpPtr = _T(0);

			if(*(lpPtr - 1) == cr) // If there is \r char before \n, set it to zero too.
				*(lpPtr - 1) = _T(0);

			lpPtr++;

			lFilePos = (LONG)_tcslen(lpPtr)*sizeof(TCHAR);

			SetFilePointer(hFile_, lFilePos*(-1), NULL, FILE_CURRENT);

			break; // If we found end of line, stop looping.
		};

	}; // while(fLoop)

	dwRetSize = (DWORD)_tcslen(lpBuf);

	*lpszBuf_ = lpBuf;

	*lpdwOutBufSize_ = dwBufSize;

	cfdprintf(_T("!cfReadLine"));

	return(dwRetSize);
};

////////////////////////////////////////////////////////////////////////
// User notification and logging functions.
////////////////////////////////////////////////////////////////////////

VOID cfAlertBox(UINT uType_, LPCTSTR szCapt_, LPCTSTR szFormat_, ...)
{
	UINT    uType;
	LPTSTR  szBuf;
	INT     nLen;
	va_list pArgs;

	if(uType_ == EVENTLOG_ERROR_TYPE)
		uType = MB_ICONHAND;
	else if(uType_ == EVENTLOG_WARNING_TYPE)
		uType = MB_ICONEXCLAMATION;
	else
		uType = MB_ICONASTERISK;

	va_start(pArgs, szFormat_);

	nLen = _vsctprintf(szFormat_, pArgs);

	szBuf = new TCHAR[nLen + 1];

	_vstprintf(szBuf, szFormat_, pArgs);

	va_end(pArgs);

	MessageBox(NULL, szBuf, szCapt_, MB_OK|uType);

	delete[] szBuf;
};

VOID cfAlertBoxError(LPCTSTR szCapt_, LPCTSTR szText_, DWORD dwErr_)
{
	HLOCAL     hLocal;

	cfGetFormatMessage(hLocal, dwErr_);

	cfAlertBox(EVENTLOG_ERROR_TYPE, szCapt_, cfEVENT_ALERT_TEXT_FORNAT, 
			   szText_, dwErr_, dwErr_, (LPTSTR)LocalLock(hLocal));

	LocalFree(hLocal);
};

VOID cfAlertBoxWarning(LPCTSTR szCapt_, LPCTSTR szText_, DWORD dwErr_)
{
	HLOCAL     hLocal;

	cfGetFormatMessage(hLocal, dwErr_);

	cfAlertBox(EVENTLOG_WARNING_TYPE, szCapt_, cfEVENT_ALERT_TEXT_FORNAT, 
			   szText_, dwErr_, dwErr_, (LPTSTR)LocalLock(hLocal));

	LocalFree(hLocal);
};

VOID cfLogEvent(LPCTSTR szSvcName_, WORD wType_, DWORD dwEvtID_, LPCTSTR szFormat_, ...)
{
	LPTSTR  szBuf = NULL;
	HANDLE  hEvtSrc;
	LPTSTR  lpszStrings[1];
	INT     nLen;
	va_list pArgs;

	if(szFormat_ != NULL)
	{
		va_start(pArgs, szFormat_);

		nLen = _vsctprintf(szFormat_, pArgs);

		szBuf = new TCHAR[nLen + 1];

		_vstprintf(szBuf, szFormat_, pArgs);

		va_end(pArgs);
	};

	lpszStrings[0] = szBuf;

	// Get a handle to use with ReportEvent(). 
	hEvtSrc = RegisterEventSource(NULL, szSvcName_);
	if(hEvtSrc != NULL)
	{
		// Write to event log.
		ReportEvent(hEvtSrc, wType_, 0, dwEvtID_, NULL, 1, 
					0, (LPCTSTR*)&lpszStrings[0], NULL);
		DeregisterEventSource(hEvtSrc);
	};
};

VOID cfLogEventError(LPCTSTR szSvcName_, DWORD dwEvtID_, LPCTSTR szText_, DWORD dwErr_)
{
	HLOCAL     hLocal;

	cfGetFormatMessage(hLocal, dwErr_);

	cfLogEvent(szSvcName_, EVENTLOG_ERROR_TYPE, dwEvtID_, cfEVENT_ALERT_TEXT_FORNAT, 
			   szText_, dwErr_, dwErr_, (LPTSTR)LocalLock(hLocal));

	LocalFree(hLocal);
};

VOID cfLogEventWarning(LPCTSTR szSvcName_, DWORD dwEvtID_, LPCTSTR szText_, DWORD dwErr_)
{
	HLOCAL     hLocal;

	cfGetFormatMessage(hLocal, dwErr_);

	cfLogEvent(szSvcName_, EVENTLOG_WARNING_TYPE, dwEvtID_, cfEVENT_ALERT_TEXT_FORNAT, 
			   szText_, dwErr_, dwErr_, (LPTSTR)LocalLock(hLocal));

	LocalFree(hLocal);
};

VOID cfInformError(BOOL fLogEvt_, LPCTSTR szCapt_, LPCTSTR szSvcName_, DWORD dwEvtID_, LPCTSTR szText_, DWORD dwErr_)
{
	if(fLogEvt_)
	{
		cfLogEventError(szSvcName_, dwEvtID_, szText_, dwErr_);
	}
	else
	{
		cfAlertBoxError(szCapt_, szText_, dwErr_);
	};
};

VOID cfInformWarning(BOOL fLogEvt_, LPCTSTR szCapt_, LPCTSTR szSvcName_, DWORD dwEvtID_, LPCTSTR szText_, DWORD dwErr_)
{
	if(fLogEvt_)
	{
		cfLogEventWarning(szSvcName_, dwEvtID_, szText_, dwErr_);
	}
	else
	{
		cfAlertBoxWarning(szCapt_, szText_, dwErr_);
	};
};

////////////////////////////////////////////////////////////////////////
// Common dialogs...
////////////////////////////////////////////////////////////////////////

BOOL ShowOpenFileDlg(HWND hWnd_, LPTSTR szFile_, LONG nFileSize_, LPTSTR szFilter_, LPTSTR szTitle_, DWORD dwFlags_, LPTSTR szDefExt_)
{
	OPENFILENAME ofn;

	ZeroMemory(&ofn, sizeof(ofn));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hWnd_;
	ofn.lpstrFile = szFile_;
	ofn.lpstrFile[0] = '\0';
	ofn.nMaxFile = nFileSize_;
	ofn.lpstrFilter = szFilter_;
	ofn.nFilterIndex = 1;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST|OFN_ENABLESIZING|OFN_LONGNAMES|OFN_EXPLORER|dwFlags_;
	ofn.lpstrTitle = szTitle_;
	ofn.lpstrDefExt = szDefExt_;
	
	return(GetOpenFileName(&ofn));
};

BOOL ShowSaveFileDlg(HWND hWnd_, LPTSTR szFile_, LONG nFileSize_, LPTSTR szFilter_, LPTSTR szTitle_, DWORD dwFlags_, LPTSTR szDefExt_)
{
	OPENFILENAME ofn;

	ZeroMemory(&ofn, sizeof(ofn));

	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hWnd_;
	ofn.lpstrFile = szFile_;
	ofn.lpstrFile[0] = '\0';
	ofn.nMaxFile = nFileSize_;
	ofn.lpstrFilter = szFilter_;
	ofn.nFilterIndex = 1;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST|OFN_ENABLESIZING|OFN_LONGNAMES|OFN_EXPLORER|OFN_OVERWRITEPROMPT|dwFlags_;
	ofn.lpstrTitle = szTitle_;
	ofn.lpstrDefExt = szDefExt_;
	
	return(GetSaveFileName(&ofn));
};

////////////////////////////////////////////////////////////////////////
// Windowing functions.
////////////////////////////////////////////////////////////////////////

LPTSTR cfListView_GetItemText(HWND hwnd_, INT nItm_, INT nSubItm_)
{
	LVITEM	lvi;
	INT		nLen;
	INT		nRes;
	BOOL	fLoop;

	memset(&lvi, 0, sizeof(LVITEM));
	lvi.iSubItem = nSubItm_;

	nLen = 128;

	fLoop = TRUE;

	while(fLoop)
	{
		nLen *= 2;
		lvi.cchTextMax = nLen;
		lvi.pszText = new TCHAR[nLen + 1];

		nRes = (INT)SendMessage(hwnd_, LVM_GETITEMTEXT, nItm_, (LPARAM)&lvi);

		if(nRes == nLen - 1)
			delete[] lvi.pszText;
		else
			fLoop = FALSE;
	};

	return(lvi.pszText);
};

LONG cfGetDlgItemLong(HWND hDlg_, UINT uID_, INT nIndex_)
{
	HWND hWnd;

	hWnd = GetDlgItem(hDlg_, uID_);
	return(GetWindowLong(hWnd, nIndex_));
};

LONG cfSetDlgItemLong(HWND hDlg_, UINT uID_, INT nIndex_, LONG lNewLong_)
{
	HWND hWnd;

	hWnd = GetDlgItem(hDlg_, uID_);
	return(SetWindowLong(hWnd, nIndex_, lNewLong_));
};

BOOL cfEnableDlgItem(HWND hDlg_, UINT uID_, BOOL fEnable_)
{
	HWND hWnd;
	hWnd = GetDlgItem(hDlg_, uID_);
	return(EnableWindow(hWnd, fEnable_));
};

INT cfGetDlgItemTextLength(HWND hDlg_, UINT uID_)
{
	HWND hWnd;
	hWnd= GetDlgItem(hDlg_, uID_);
	return(GetWindowTextLength(hWnd));
};

////////////////////////////////////////////////////////////////////////
// Registry functions.
////////////////////////////////////////////////////////////////////////

INT cfRegReadDWVal(HKEY hKey_, LPTSTR szName_, 
						LPDWORD pdwVal_, DWORD dwDef_)
{
	LONG  lRet;
	DWORD dwType;
	DWORD dwSize = sizeof(DWORD);

	lRet = RegQueryValueEx(hKey_, szName_, NULL, &dwType,
						   (LPBYTE)pdwVal_, &dwSize);

	if(lRet != ERROR_SUCCESS) 
		*pdwVal_ = dwDef_;

	return lRet;
};

INT cfRegReadSTRVal(HKEY hKey_, LPTSTR szName_, 
						 LPTSTR* szVal_, LPTSTR szDef_)
{
	DWORD dwType, dwSize;
	LONG lRet;

	lRet = RegQueryValueEx(hKey_, szName_, NULL, &dwType,
						   (LPBYTE)NULL, &dwSize);

	if(szVal_ != NULL)
	{
		if(lRet == ERROR_SUCCESS)
		{
			*szVal_ = new TCHAR[dwSize+1];
			RegQueryValueEx(hKey_, szName_, NULL, &dwType,
							(LPBYTE)*szVal_, &dwSize);
		}
		else
		{
			if(szDef_ != NULL)
			{
				*szVal_ = new TCHAR[_tcslen(szDef_)+1];
				_tcscpy(*szVal_, szDef_);
			}
			else
			{
				*szVal_ = NULL;
			};
		};
	};

	return lRet;
};

INT cfRegSaveDWVal(HKEY hKey_, LPTSTR szName_, DWORD dwVal_)
{
	LONG lRet;

	lRet = RegSetValueEx(hKey_, szName_, 0, REG_DWORD,
						 (LPBYTE)&dwVal_, sizeof(DWORD));

	return lRet;
};

INT cfRegSaveSTRVal(HKEY hKey_, LPTSTR szName_, LPTSTR szVal_)
{
	LONG lRet;

	lRet = RegSetValueEx(hKey_, szName_, 0, REG_SZ,
						 (LPBYTE)szVal_, (DWORD)_tcslen(szVal_)+1);

	return lRet;
};