/*  Botik::Key Pro - Botik Key Client for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2004 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "stdafx.h"
#include "botikkeypro.h"
#include "service/eventlog.h"
#include "main_dlg.h"
#include "interface_h.h"
#include <mmsystem.h>
#include "logviewer.h"

//////////////////////////////////////////////////////////////////////////////
// Global variables.
//////////////////////////////////////////////////////////////////////////////
HICON		g_hBigIconLV = NULL;
HICON		g_hSmallIconLV = NULL;

HIMAGELIST	g_hImgLstTypes = NULL;
HIMAGELIST	g_hImgLstNorm = NULL;
HIMAGELIST	g_hImgLstHot = NULL;
HIMAGELIST	g_hImgLstDis = NULL;

HWND		g_hwndLogViewer = NULL;
HWND		g_hwndListView = NULL;
HWND		g_hwndToolbar = NULL;
HWND		g_hwndRebar = NULL;
HWND		g_hwndStatusbar = NULL;

DWORD		g_dwWinCxLV = 0;
DWORD		g_dwWinCyLV = 0;
BOOL		g_fIsMaximized = FALSE;
BOOL		g_fStatusbar = TRUE;
BOOL		g_fToolbar = TRUE;
BOOL		g_fSvcSync = TRUE;
BOOL		g_fAutoscroll = TRUE;
DWORD		g_dwFirstColWidth = 120;
DWORD		g_dwSecondColWidth = 300;
DWORD		g_dwMaxEventsShow = 1000;
BOOL		g_fPlaySound;

HANDLE		g_hClipTextLV = NULL;
HANDLE		g_hClipOEMLV = NULL;
HANDLE		g_hClipUTFLV = NULL;

//////////////////////////////////////////////////////////////////////////////
// Forward declarations.
//////////////////////////////////////////////////////////////////////////////

LRESULT CALLBACK WndProcLV(HWND hWnd_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_);
LRESULT CALLBACK WndProcPrgDlg(HWND hDlg_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_);
LRESULT OnCommandLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_);
LRESULT OnCreateLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_);
LRESULT OnSizeLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_);
LRESULT OnNotifyLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_);
LRESULT OnInitGUIMenuLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_);
LRESULT OnNewEvent(HWND hWnd_, WPARAM wParam_, LPARAM lParam_);

HWND CreateListViewLV(HWND hPrn_);
HWND CreateToolbarLV(HWND hPrn_);
BOOL CreateRebarLV(HWND hPrn_);
BOOL CreateImageListsLV();
HWND CreateStatusbarLV(HWND hPrn_);
VOID ShowToolbarLV(BOOL fShow_);
BOOL AddToolbarButtonLV(HWND hwndTlb_, DWORD dwID_, DWORD dwImID_, DWORD dwData_, BYTE bStyle_, BYTE bState_ = TBSTATE_ENABLED);
BOOL AddImageIntoImageListLV(HIMAGELIST hImgLst_, DWORD dwImgID_, DWORD dwImgIDMask_);
VOID ShowStatusbarLV(BOOL fShow_);
VOID ShowToolbarLV(BOOL fShow_);
BOOL PutInClipboard(LPTSTR szStr_);
VOID SaveFile(HWND hWnd_);
VOID OpenFile(HWND hWnd_);
VOID PutSelectedInClipboard(HWND hWnd_);
VOID ParseLogFileLine(LPTSTR szLine_);

//////////////////////////////////////////////////////////////////////////////
// Implementation.
//////////////////////////////////////////////////////////////////////////////

HWND CreateLogViewerWnd(HWND hPrnWnd_)
{
	g_hwndLogViewer = CreateWindow(ABK_WNDCLS_LOG_VIEWER, ABK_LOGVIEWER_TITLE, 
				WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, g_dwWinCxLV, g_dwWinCyLV, 
				NULL, NULL, g_hInst, NULL);

	if(!g_hwndLogViewer)
	{
		cfdprintf(_T("CreateLogViewerWnd: Failed to create Log Viewer window (%d)"), GetLastError());
		return(NULL);
	};

	UpdateWindow(g_hwndLogViewer);

	return(g_hwndLogViewer);
};

BOOL RegisterLogViewerWndCls()
{
	WNDCLASSEX	wcex;

	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)WndProcLV;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= g_hInst;
	wcex.hIcon			= LoadIcon(g_hInst, (LPCTSTR)IDI_LOGVIEWER);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_BTNFACE+1);//(COLOR_WINDOW+1);
	wcex.lpszMenuName	= (LPCTSTR)IDC_LOGVIEWER;
	wcex.lpszClassName	= ABK_WNDCLS_LOG_VIEWER;
	wcex.hIconSm		= (HICON)LoadImage(g_hInst, MAKEINTRESOURCE(IDI_LOGVIEWER), IMAGE_ICON, 16, 16, 0);

	if(!RegisterClassEx(&wcex))
	{
		cfdprintf(_T("RegisterLogViewerWndCls: Failed to register window class (%d)"), GetLastError());
		return(FALSE);
	};

	return(TRUE);
};

LRESULT CALLBACK WndProcLV(HWND hWnd_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_)
{
	PAINTSTRUCT ps;
	HDC hdc;

	switch (uMsg_) 
	{
	case WM_CREATE:
		return(OnCreateLV(hWnd_, wParam_, lParam_));

	case WM_DESTROYCLIPBOARD :
		cfdprintf(_T("WM_DESTROYCLIPBOARD"));
		if(g_hClipTextLV)
		{
			GlobalFree(g_hClipTextLV);
			g_hClipTextLV = NULL;
		};

		if(g_hClipOEMLV)
		{
			GlobalFree(g_hClipOEMLV);
			g_hClipOEMLV = NULL;
		};

		if(g_hClipUTFLV)
		{
			GlobalFree(g_hClipUTFLV);
			g_hClipUTFLV = NULL;
		};
		break;

	case WM_INITMENUPOPUP:
		OnInitGUIMenuLV(hWnd_, wParam_, lParam_);
		break;

	case WM_COMMAND:
		OnCommandLV(hWnd_, wParam_, lParam_);
		break;

	case WM_SIZE:
		OnSizeLV(hWnd_, wParam_, lParam_);
		break;

	case WM_PAINT:
		hdc = BeginPaint(hWnd_, &ps);
		// TODO: Add any drawing code here...
		EndPaint(hWnd_, &ps);
		break;

	case WM_NOTIFY :
		return(OnNotifyLV(hWnd_, wParam_, lParam_));

	case WM_DESTROY:
		g_dwFirstColWidth = ListView_GetColumnWidth(g_hwndListView, 1);
		g_dwSecondColWidth = ListView_GetColumnWidth(g_hwndListView, 2);

		if(!g_fIsMaximized && !IsMinimized(hWnd_))
		{
			RECT rc;

			GetWindowRect(hWnd_, &rc);

			g_dwWinCxLV = rc.right - rc.left;
			g_dwWinCyLV = rc.bottom - rc.top;
		};
		break;

	case WM_USER_NEW_EVENT:
		if(g_fSvcSync)
			OnNewEvent(hWnd_, wParam_, lParam_);
		break;

	case WM_SYSCOMMAND:
		if(wParam_ == SC_CLOSE)
			ShowWindow(hWnd_, SW_HIDE);
		else
		{
			if(wParam_ == SC_MAXIMIZE)
				g_fIsMaximized = TRUE;
			else if(wParam_ == SC_RESTORE)
				g_fIsMaximized = FALSE;

			DefWindowProc(hWnd_, uMsg_, wParam_, lParam_);
		};
		break;

	default:
		return(DefWindowProc(hWnd_, uMsg_, wParam_, lParam_));
	};

	return(0);
};

// Message handler proc for progress box.
LRESULT CALLBACK WndProcPrgDlg(HWND hDlg_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_)
{
	switch (uMsg_)
	{
	case WM_INITDIALOG:
		{
			//HRSRC hrc;
			//HGLOBAL hgl;
			//LPTSTR szStr;

			//hrc = FindResource((HMODULE)g_hInst, MAKEINTRESOURCE(IDR_GPL_LICENSE), RT_HTML);
			//if(hrc == NULL)
			//	return(FALSE);

			//hgl = LoadResource((HMODULE)g_hInst, hrc);
			//if(hgl == NULL)
			//	return(FALSE);

			//szStr = (LPTSTR)LockResource(hgl);
			//if(szStr == NULL)
			//	return(FALSE);

			//SetDlgItemText(hDlg_, IDC_LICENSE_TEXT, szStr);
		};
		return(TRUE);

	case WM_ACTIVATE:
		cfdprintf(_T("WM_ACTIVATE"));
		if(wParam_ == WA_INACTIVE && (HWND)lParam_ == g_hwndLogViewer)
		{
			cfdprintf(_T("Set foreground."));
			if(SetActiveWindow(hDlg_) == NULL)
				cfdprintf(_T("Err: %d"), GetLastError());

			if(!SetForegroundWindow(hDlg_))
				cfdprintf(_T("Err: %d"), GetLastError());
		};
		break;

	case WM_COMMAND:
		if (LOWORD(wParam_) == IDOK || LOWORD(wParam_) == IDCANCEL) 
		{
			EndDialog(hDlg_, LOWORD(wParam_));
			return(TRUE);
		};
		break;
	};

	return(FALSE);
};

VOID ShowLogViewerWindow()
{
	if(g_fIsMaximized)
		ShowWindow(g_hwndLogViewer, SW_SHOWMAXIMIZED);
	else
		ShowWindow(g_hwndLogViewer, SW_SHOW);

	SetForegroundWindow(g_hwndLogViewer);
	UpdateWindow(g_hwndLogViewer);
};

//////////////////////////////////////////////////////////////////////////////
// Message handles.
//////////////////////////////////////////////////////////////////////////////

LRESULT OnNewEvent(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	DWORD          dwEvtType;
	BOOL           fPlayErr = FALSE;
	BOOL           fPlayMsg = FALSE;
	static BOOL    fErrPlayed = FALSE;
	SYSTEMTIME     stTime;
	unsigned char* szMsg = NULL; // _MUST_ set to NULL in order to allocate memory buffer by the client stub.

	cfdprintf(_T("OnNewEvent: WM_USER_NEW_EVENT"));

	// Pull all stored events.
	while(SvcPullEvent((UINT*)&dwEvtType, &stTime, &szMsg))
	{
		cfdprintf(_T("OnNewEvent: %d, %s, %d:%d:%d"), dwEvtType, szMsg, stTime.wHour, stTime.wMinute, stTime.wSecond);

		// XXX UNICODE uncompatible!
		AddListViewLineLV(dwEvtType, &stTime, (LPTSTR)szMsg);

		if(   g_fForceLogviewer 
		   && (   dwEvtType == BKP_EVENT_TYPE_ERROR 
		       || dwEvtType == BKP_EVENT_TYPE_MESSAGE))
		{
			ShowLogViewerWindow();
		};

		// If last message was error, do not play error.wav again.
		if(dwEvtType != BKP_EVENT_TYPE_ERROR)
			fErrPlayed = FALSE;

		if(dwEvtType == BKP_EVENT_TYPE_ERROR && !fErrPlayed)
		{
			fPlayErr = TRUE;
		};

		if(dwEvtType == BKP_EVENT_TYPE_MESSAGE)
			fPlayMsg = TRUE;

		midl_user_free(szMsg); // Free memory buffer allocated by the client stub.

		szMsg = NULL;
	};

	if(fPlayErr && g_fPlaySound)
	{
		PlaySound(_T("error.wav"), NULL, SND_FILENAME|SND_ASYNC);
		fErrPlayed = TRUE;
	};

	if(fPlayMsg && g_fPlaySound)
		PlaySound(_T("message.wav"), NULL, SND_FILENAME|SND_ASYNC);

	return(0);
};

LRESULT OnInitGUIMenuLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	MENUITEMINFO	mii;
	DWORD			dwPos;
	BOOL			fIsWin;
	HMENU			hMenu;
	INT				nItmCnt;

	dwPos = LOWORD(lParam_);
	fIsWin = HIWORD(lParam_);

	hMenu = (HMENU)wParam_;

	mii.cbSize = sizeof(MENUITEMINFO);
	mii.fMask = MIIM_STATE;

	cfdprintf(_T("OnInitGUIMenu: %d, %d"), fIsWin, g_fStatusbar);

	nItmCnt = ListView_GetItemCount(g_hwndListView);

	if(fIsWin == FALSE && dwPos == 0)
	{
		mii.fState = nItmCnt ? MFS_ENABLED : MFS_DISABLED;
		SetMenuItemInfo(hMenu, 0, TRUE, &mii);

		mii.fState = g_fSvcSync ? MFS_DISABLED : MFS_ENABLED;
		SetMenuItemInfo(hMenu, 3, TRUE, &mii);
	}
	else if(fIsWin == FALSE && dwPos == 1)
	{
		mii.fState = nItmCnt ? MFS_ENABLED : MFS_DISABLED;
		SetMenuItemInfo(hMenu, 0, TRUE, &mii);
		SetMenuItemInfo(hMenu, 3, TRUE, &mii);

		mii.fState = nItmCnt ? MFS_DISABLED : MFS_DISABLED;
		SetMenuItemInfo(hMenu, 4, TRUE, &mii);

		nItmCnt = ListView_GetSelectedCount(g_hwndListView);
		mii.fState = nItmCnt ? MFS_ENABLED : MFS_DISABLED;
		SetMenuItemInfo(hMenu, 1, TRUE, &mii);
	} 
	else if(fIsWin == FALSE && dwPos == 2)
	{
		mii.fState = g_fToolbar ? MFS_CHECKED : MFS_UNCHECKED;
		SetMenuItemInfo(hMenu, 0, TRUE, &mii);

		mii.fState = g_fStatusbar ? MFS_CHECKED : MFS_UNCHECKED;
		SetMenuItemInfo(hMenu, 1, TRUE, &mii);

		mii.fState = g_fAutoscroll ? MFS_CHECKED : MFS_UNCHECKED;
		SetMenuItemInfo(hMenu, 3, TRUE, &mii);
	};

	DefWindowProc(hWnd_, WM_INITMENUPOPUP, wParam_, lParam_);

	return(0);
};

LRESULT OnNotifyLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	LPNMHDR lpnmh = (LPNMHDR)lParam_;
	LONG	lRes = 0;

	if(lpnmh->hwndFrom == g_hwndListView)
	{
		if(lpnmh->code == LVN_ITEMCHANGED)
		{
			INT nSelCnt;

			nSelCnt = ListView_GetSelectedCount(g_hwndListView);

			if(nSelCnt)
				SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0003, MAKELONG(TBSTATE_ENABLED, 0));
			else
				SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0003, MAKELONG(TBSTATE_INDETERMINATE, 0));
		}
		else if(lpnmh->code == LVN_INSERTITEM)
		{
			SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0000, MAKELONG(TBSTATE_ENABLED, 0));
			SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0002, MAKELONG(TBSTATE_ENABLED, 0));
			SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0005, MAKELONG(TBSTATE_ENABLED, 0));
			//SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0006, MAKELONG(TBSTATE_ENABLED, 0));
		}
		else if(   lpnmh->code == LVN_DELETEALLITEMS 
				/*|| lpnmh->code == LVN_DELETEITEM*/)
		{
			INT nItmCnt;

			cfdprintf(_T("LVN_DELETEITEM"));

			nItmCnt = ListView_GetItemCount(g_hwndListView);

			//if(nItmCnt == 0)
			//{
				SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0000, MAKELONG(TBSTATE_INDETERMINATE, 0));
				SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0002, MAKELONG(TBSTATE_INDETERMINATE, 0));
				SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0005, MAKELONG(TBSTATE_INDETERMINATE, 0));
				SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0006, MAKELONG(TBSTATE_INDETERMINATE, 0));
			//};

			lRes = TRUE; // To prevent LVN_DELETEITEM if LVN_DELETEALLITEMS received.
		}
		else if(lpnmh->code == LVN_DELETEITEM)
		{
			RECT rc;

			// XXX This is a hack to scroll free space after the last column in the list, if any.
			GetClientRect(g_hwndListView, &rc);
			InvalidateRect(g_hwndListView, &rc, FALSE);
		}
		else if(lpnmh->code == NM_CUSTOMDRAW)
		{
			LPNMLVCUSTOMDRAW lpnmlvcd = (LPNMLVCUSTOMDRAW)lpnmh;

			switch(lpnmlvcd->nmcd.dwDrawStage)
			{
			case CDDS_PREPAINT:
				lRes |= CDRF_NOTIFYITEMDRAW/*|CDRF_SKIPDEFAULT*/;
				break;

			case CDDS_ITEMPREPAINT:
				{
					HDC			hdc;
					HDC			hdci;
					RECT		rcib;
					RECT		rcbg;
					RECT		rc;
					DWORD		nItem;
					LV_ITEM		lvi;
					HRGN		hRgn;

					hdci = lpnmlvcd->nmcd.hdc;
					nItem = (DWORD)lpnmlvcd->nmcd.dwItemSpec;

					// Get item data and rects.
					lvi.mask = LVIF_IMAGE|LVIF_STATE|LVIF_PARAM;
					lvi.iItem = nItem;
					lvi.iSubItem = 0;
					lvi.stateMask = 0xFFFF;

					ListView_GetItem(g_hwndListView, &lvi);
					ListView_GetItemRect(g_hwndListView, nItem, &rcib, LVIR_BOUNDS);

					rcbg = rcib;

					GetClientRect(g_hwndListView, &rc);

					// If row is less than window width, set window width.
					if(rcbg.right < rc.right)
						rcbg.right = rc.right;

					// Create my DC with my clip rgn.
					RECT rcbr;
					
					GetClipBox(hdci, &rcbr);
					rcbr.right = rc.right;
					hRgn = CreateRectRgnIndirect(&rcbg);
					hdc = GetDC(g_hwndListView);
					SelectClipRgn(hdc, hRgn);
					DeleteObject(hRgn);

					SelectObject(hdc, GetCurrentObject(hdci, OBJ_FONT));

					////////////////////
					// Fill item backgroud...
					COLORREF	crBg;
					COLORREF	crFg;
					HBRUSH		hBrBg;

					crFg = ABK_NORMAL_FGCOLOR;

					if(lvi.state & LVIS_SELECTED)
					{
						crBg = ABK_SELECTED_BGCOLOR;
						crFg = ABK_SELECTED_FGCOLOR;
					}
					else if(lvi.lParam == BKP_EVENT_TYPE_ERROR)
						crBg = ABK_ERROR_BGCOLOR;
					else if(lvi.lParam == BKP_EVENT_TYPE_INFORMATION)
						crBg = ABK_INFORM_BGCOLOR;
					else if(lvi.lParam == BKP_EVENT_TYPE_OK)
						crBg = ABK_OK_BGCOLOR;
					else if(lvi.lParam == BKP_EVENT_TYPE_MESSAGE)
						crBg = ABK_MSG_BGCOLOR;

					hBrBg = CreateSolidBrush(crBg);

					FillRect(hdc, &rcbg, hBrBg);

					DeleteObject(hBrBg);

					SetTextColor(hdc, crFg);
					SetBkMode(hdc, TRANSPARENT);

					///////////////////////////
					// Draw state image if any.
					HIMAGELIST	hImgLst;
					RECT		rci;

					hImgLst = ListView_GetImageList(g_hwndListView, LVSIL_SMALL);
					ListView_GetItemRect(g_hwndListView, nItem, &rci, LVIR_ICON);

					// Set clip region.
					rci.right = rcbg.left + ListView_GetColumnWidth(g_hwndListView, 0);
					hRgn = CreateRectRgnIndirect(&rci);
					SelectClipRgn(hdc, hRgn);
					DeleteObject(hRgn);

					if(lvi.iImage != -1)
					{
						ImageList_Draw(hImgLst, lvi.iImage, hdc, 
							rci.left, rci.top, ILD_TRANSPARENT|((lvi.state & LVIS_SELECTED)?ILD_BLEND25:0));
					};

					SelectClipRgn(hdc, NULL);

					///////////////////////
					// Print out item text.
					RECT		rcol;
					INT			nCol;
					LVCOLUMN	lvc;
					INT			offset;

					rcol = rcib;
					rcol.right = rcol.left + ListView_GetColumnWidth(g_hwndListView, 0);

					offset = rci.left - rcol.left; // same offset as in the first col.

					lvc.mask = LVCF_WIDTH/* | LVCF_FMT*/;

					for(nCol = 1; ListView_GetColumn(g_hwndListView, nCol, &lvc); nCol++)
					{
						LPTSTR	szTxt;

						rcol.left = rcol.right + offset;
						rcol.right = rcol.left + lvc.cx - offset;

						szTxt = cfListView_GetItemText(g_hwndListView, nItem, nCol);

						if(_tcslen(szTxt) == 0)
							continue;

						// DrawFocusRect(hdc, &rcol);

						DrawText(hdc, szTxt, -1, &rcol, 
							DT_SINGLELINE|DT_NOPREFIX|DT_VCENTER|DT_END_ELLIPSIS);

						delete[] szTxt;
					};

					ReleaseDC(g_hwndListView, hdc);

					lRes |= CDRF_SKIPDEFAULT;
				};
				break; // case CDDS_ITEMPREPAINT
			}; // switch(lpnmlvcd->nmcd.dwDrawStage)
		}; // if(lpnmh->code == NM_CUSTOMDRAW)
	}; // if(lpnmh->hwndFrom == g_hwndListView)

	return(lRes);
}; // LRESULT OnNotifyLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)

LRESULT OnSizeLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	RECT rc, rcrb, rcsb;

	SendMessage(g_hwndStatusbar, WM_SIZE, wParam_, lParam_);

	GetClientRect(hWnd_, &rc);
	GetWindowRect(g_hwndRebar, &rcrb);
	GetWindowRect(g_hwndStatusbar, &rcsb);

	MoveWindow(g_hwndRebar, rc.left-2, rc.top, rc.right-rc.left+4, rcrb.top-rcrb.bottom, TRUE);

	rc.top += rcrb.bottom-rcrb.top;

	if(g_fStatusbar)
		rc.bottom -= rcsb.bottom-rcsb.top;

	MoveWindow(g_hwndListView, rc.left, rc.top, rc.right-rc.left, rc.bottom-rc.top, TRUE);

	return(0);
};

LRESULT OnCreateLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	DWORD dwStyle;

	if(!CreateImageListsLV())
		return(0);

	if(!CreateRebarLV(hWnd_))
		return(0);

	AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_0000, 0, 0, BTNS_BUTTON|BTNS_AUTOSIZE, TBSTATE_INDETERMINATE);
	AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_SEP_0000, I_IMAGENONE, 0, BTNS_SEP);
	AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_0001, 1, 0, BTNS_BUTTON|BTNS_AUTOSIZE);
	AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_0004, 4, 0, BTNS_BUTTON|BTNS_AUTOSIZE, TBSTATE_INDETERMINATE);
	AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_SEP_0001, I_IMAGENONE, 0, BTNS_SEP);
	AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_0002, 2, 0, BTNS_BUTTON|BTNS_AUTOSIZE, TBSTATE_INDETERMINATE);
	AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_0003, 3, 0, BTNS_BUTTON|BTNS_AUTOSIZE, TBSTATE_INDETERMINATE);
	AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_SEP_0002, I_IMAGENONE, 0, BTNS_SEP);
	AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_0007, 7, 0, BTNS_BUTTON|BTNS_AUTOSIZE|TBSTYLE_CHECK);
	//AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_SEP_0003, I_IMAGENONE, 0, BTNS_SEP);
	//AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_0005, 5, 0, BTNS_BUTTON|BTNS_AUTOSIZE, TBSTATE_INDETERMINATE);
	//AddToolbarButtonLV(g_hwndToolbar, IDS_BUTTON_0006, 6, 0, BTNS_BUTTON|BTNS_AUTOSIZE, TBSTATE_INDETERMINATE);

	if(!CreateStatusbarLV(hWnd_))
		return(0);

	if(!CreateListViewLV(hWnd_))
		return(0);

	ShowToolbarLV(g_fToolbar);

	dwStyle = g_fAutoscroll ? TBSTATE_CHECKED : 0;

	SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0007, MAKELONG(dwStyle|TBSTATE_ENABLED, 0));

	return(0);
};

LRESULT OnCommandLV(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	INT nId    = LOWORD(wParam_); 
	INT nEvent = HIWORD(wParam_); 

	// Parse the menu selections:
	switch (nId)
	{
	case IDM_LV_EDIT_DEL:
	case IDS_BUTTON_0002:
		{
			INT nRet;

			nRet = MessageBox(hWnd_, _T("    ?"), 
						ABK_LOGVIEWER_TITLE,
						MB_YESNO|MB_ICONQUESTION);

			if(nRet == IDYES)
			{
				ListView_DeleteAllItems(g_hwndListView);
			};
		};
		break;

	case IDM_LV_VIEW_AUTOSCROLL:
		{
			DWORD dwStyle = !g_fAutoscroll ? TBSTATE_CHECKED : 0;

			SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0007, MAKELONG(dwStyle|TBSTATE_ENABLED, 0));
		};
	case IDS_BUTTON_0007:
		g_fAutoscroll = !g_fAutoscroll;
		break;

	case IDM_LV_FILE_SAVE:
	case IDS_BUTTON_0000:
		SaveFile(hWnd_);
		break;

	case IDM_LV_FILE_OPEN:
	case IDS_BUTTON_0001:
		OpenFile(hWnd_);
		break;

	case IDM_LV_FILE_SYNC:
	case IDS_BUTTON_0004:
		g_fSvcSync = TRUE;
		ListView_DeleteAllItems(g_hwndListView);
		SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0004, MAKELONG(TBSTATE_INDETERMINATE, 0));
		break;

	case IDM_LV_EDIT_COPY:
	case IDS_BUTTON_0003:
		PutSelectedInClipboard(hWnd_);
		break;

	case IDM_LV_VIEW_TOOLBAR:
		ShowToolbarLV(g_fToolbar = !g_fToolbar);
		break;

	case IDM_LV_VIEW_STATUSBAR:
		ShowStatusbarLV(g_fStatusbar = !g_fStatusbar);
		break;

	case IDM_LV_HELP_HELP:
		{
			LPTSTR  szPath;
			LPCTSTR szTopic = _T("::/html/02_logviewer.html");

			szPath = new TCHAR[_tcslen(g_szChmPath) + _tcslen(szTopic) + 1];

			_tcscpy(szPath, g_szChmPath);
			_tcscat(szPath, _T(szTopic));

			HtmlHelp(HWND_DESKTOP/*g_hMainDlg*/, szPath, HH_DISPLAY_TOPIC, NULL);

			delete[] szPath;
		};
		break;

	case IDM_LV_HELP_SITE:
		ShellExecute(HWND_DESKTOP, _T("open"), ABK_DEV_LINK_SITE, NULL, NULL, SW_SHOWNORMAL);
		break;

	case IDM_LV_HELP_GPL:
		DialogBox(g_hInst, MAKEINTRESOURCE(IDD_LICENSE_DLG), hWnd_, (DLGPROC)LicenseDlg);
		break;

	case IDM_LV_HELP_ABOUT:
		DialogBox(g_hInst, MAKEINTRESOURCE(IDD_ABOUT_DLG), hWnd_, (DLGPROC)AboutDlg);
		break;

	case IDM_LV_FILE_EXIT:
	case IDM_EXIT:
		DestroyWindow(hWnd_);
		break;

	default:
		return(DefWindowProc(hWnd_, WM_COMMAND, wParam_, lParam_));
	};

	return(0);
};

//////////////////////////////////////////////////////////////////////////////
// Help functions.
//////////////////////////////////////////////////////////////////////////////

VOID PutSelectedInClipboard(HWND hWnd_)
{
	DWORD	dwCnt;

	cfdprintf(_T("PutSelectedInClipboard"));

	dwCnt = ListView_GetSelectedCount(g_hwndListView);

	if(dwCnt)
	{
		DWORD	dwAllocDelta = 256;	// Buffer allocation delta.
		DWORD	dwBufSize = 512;	// Initial size of buffer, item size is TCHAR.
		LVITEM	lvi;
		LPTSTR	szBuf;
		LPTSTR	szTmp;
		DWORD	i = 0;

		memset(&lvi, 0, sizeof(LVITEM));

		szBuf = new TCHAR[dwBufSize + 1];
		szBuf[0] = _T(0);

		dwCnt = ListView_GetItemCount(g_hwndListView);

		while(i < dwCnt)
		{
			lvi.mask = LVIF_PARAM|LVIF_STATE;
			lvi.iItem = i;
			lvi.stateMask = LVIS_SELECTED;
			ListView_GetItem(g_hwndListView, &lvi);

			//cfBREAK_PT

			if(lvi.state == LVIS_SELECTED)
			{
				cfdprintf(_T("PutSelectedInClipboard: Item: %d"), i);

				// Get and put text describing type of message.
				szTmp = EventLogTypeToString((DWORD)lvi.lParam);

				cfdprintf(_T("PutSelectedInClipboard: EntryType: %s"), szTmp);

				// Check available size of the buffer, if it less
				// than size of the new string, resize it.
				if((dwBufSize - _tcslen(szBuf)) < (_tcslen(szTmp) + 2)) // We should add tab on the next step.
				{
					DWORD dwNewSize = dwBufSize + _tcslen(szTmp) + dwAllocDelta + 2; // for tab and for some case.

					szBuf = (LPTSTR)cfRealloc(szBuf, dwBufSize*sizeof(TCHAR), dwNewSize*sizeof(TCHAR));
					dwBufSize = dwNewSize;
				};

				_tcscat(szBuf, szTmp);

				// Get texts from the second and third columns...
				for(int j = 1; j < 3; j++)
				{
					_tcscat(szBuf, _T("\t"));

					// Get and put text from the j-column of the list control.
					szTmp = cfListView_GetItemText(g_hwndListView, i, j);

					cfdprintf(_T("PutSelectedInClipboard: Col: %d; %s"), j, szTmp);

					// Check available size of the buffer, if it less
					// than size of the new string, resize it.
					if((dwBufSize - _tcslen(szBuf)) < (_tcslen(szTmp) + 3)) // We should add tab or \r\n on the next step.
					{
						DWORD dwNewSize = dwBufSize + _tcslen(szTmp) + dwAllocDelta + 3; // for tab or \r\n.

						szBuf = (LPTSTR)cfRealloc(szBuf, dwBufSize*sizeof(TCHAR), dwNewSize*sizeof(TCHAR));
						dwBufSize = dwNewSize;
					};

					_tcscat(szBuf, szTmp);
					delete[] szTmp; // Free memory allocated in the cfListView_GetItemText.
				};

				_tcscat(szBuf, _T("\r\n"));
			}; // if(lvi.state == LVIS_SELECTED)

			i++;
		}; // while(i < dwCnt) dwCnt is a number of items in the list ctrl.

		PutInClipboard(szBuf);

		delete[] szBuf; // Free buffer.
	}; // if(dwCnt) dwCnt is a number of selected items.

	cfdprintf(_T("!PutSelectedInClipboard"));
};

VOID OpenFile(HWND hWnd_)
{
	TCHAR szFile[MAX_PATH+1];
	TCHAR szFilter[MAX_LOADSTRING+1];

	ZeroMemory(szFilter, sizeof(szFilter));
	LoadString(g_hInst, IDS_FILE_FILTER_01, szFilter, MAX_LOADSTRING);

	if(ShowOpenFileDlg(hWnd_, szFile, MAX_PATH, szFilter, 
			_T("  ..."), 0, _T("log")))
	{
		HANDLE hFile;

		hFile = CreateFile(szFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, 
							OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile == INVALID_HANDLE_VALUE)
		{
			// XXX Localization.
            cfAlertBoxError(_T("Botik::Key Pro..."),
							_T("      ."), 
							GetLastError());
		}
		else
		{
			DWORD dwOutBufSize = 0;
			DWORD dwBufSize = 0;
			DWORD dwRetSize;
			LPTSTR lpBuf = NULL;

			// It's a right place to cleanup old data stored in the list control,
			// and change a state of some buttons and states...
			g_fSvcSync = FALSE;
			ListView_DeleteAllItems(g_hwndListView);
			SendMessage(g_hwndToolbar, TB_SETSTATE, IDS_BUTTON_0004, MAKELONG(TBSTATE_ENABLED, 0));

			// File read loop...
			while(1)
			{
				dwRetSize = cfReadLine(hFile, &lpBuf, dwBufSize, &dwOutBufSize);

				dwBufSize = dwOutBufSize;

				if(dwRetSize == 0)
					break;

				ParseLogFileLine(lpBuf);

				cfdprintf(_T("lpBuf: %s; Size: %d; dwRetSize: %d"), lpBuf, dwOutBufSize, dwRetSize);
			};

			delete[] lpBuf; // Free buffer allocated by the cfReadLine.

			CloseHandle(hFile);

			// XXX. Fix it later. Update window client rect... 
			RECT rc;
			GetClientRect(g_hwndListView, &rc);
			InvalidateRect(g_hwndListView, &rc, TRUE);
		};
	};
};

VOID ParseLogFileLine(LPTSTR szLine_)
{
	LPTSTR	lpPtr;
	LPTSTR	lpTime;
	DWORD	dwType = 0;

	// Find fistr \t.
	lpPtr = _tcschr(szLine_, _T('\t'));
	if(!lpPtr) return;

	*lpPtr = _T(0);
	lpPtr++;

	if(!_tcscmp(szLine_, BKP_EVENT_TYPE_STR_INFORMATION))
		dwType = BKP_EVENT_TYPE_INFORMATION;
	else if(!_tcscmp(szLine_, BKP_EVENT_TYPE_STR_OK))
		dwType = BKP_EVENT_TYPE_OK;
	else if(!_tcscmp(szLine_, BKP_EVENT_TYPE_STR_ERROR))
		dwType = BKP_EVENT_TYPE_ERROR;
	else if(!_tcscmp(szLine_, BKP_EVENT_TYPE_STR_MESSAGE))
		dwType = BKP_EVENT_TYPE_MESSAGE;

	if(!dwType) return;

	lpTime = lpPtr;

	// Next \t.
	lpPtr = _tcschr(lpTime, _T('\t'));
	if(!lpPtr) return;

	*lpPtr = _T(0);
	lpPtr++;

	AddListViewLineLV(dwType, lpTime, lpPtr);
};

VOID SaveFile(HWND hWnd_)
{
	TCHAR szFile[MAX_PATH+1];
	TCHAR szFilter[MAX_LOADSTRING+1];

	ZeroMemory(szFilter, sizeof(szFilter));
	LoadString(g_hInst, IDS_FILE_FILTER_01, szFilter, MAX_LOADSTRING);

	if(ShowSaveFileDlg(hWnd_, szFile, MAX_PATH, szFilter, 
			_T("  ..."), 0, _T("log")))
	{
		HANDLE hFile;
		
		hFile = CreateFile(szFile, GENERIC_WRITE, FILE_SHARE_READ, NULL, 
							CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile == INVALID_HANDLE_VALUE)
		{
			// XXX Localization.
            cfAlertBoxError(_T("Botik::Key Pro..."),
							_T("      ."), 
							GetLastError());
		}
		else
		{

//			HWND	hDlg;
			DWORD	dwSavedSz;
			LPTSTR	szPtr;
			INT		nCnt;

			//hDlg = CreateDialog(g_hInst, MAKEINTRESOURCE(IDD_PROGRESS_DLG), hWnd_, (DLGPROC)WndProcPrgDlg);
			//ShowWindow(hDlg, SW_SHOW);

			nCnt = ListView_GetItemCount(g_hwndListView);

			for(INT i = 0; i < nCnt; i++)
			{
				LVITEM lvi;

				memset(&lvi, 0, sizeof(LVITEM));

				lvi.iItem = i;
				lvi.mask = LVIF_PARAM;

				ListView_GetItem(g_hwndListView, &lvi);

				szPtr = EventLogTypeToString((DWORD)lvi.lParam);

				WriteFile(hFile, szPtr, (DWORD)_tcslen(szPtr), &dwSavedSz, NULL);
				WriteFile(hFile, ABK_LOGFILE_ENTRY_DELEMITER, 
					(DWORD)_tcslen(ABK_LOGFILE_ENTRY_DELEMITER), &dwSavedSz, NULL);

				szPtr = cfListView_GetItemText(g_hwndListView, i, 1);

				WriteFile(hFile, szPtr, (DWORD)_tcslen(szPtr), &dwSavedSz, NULL);
				WriteFile(hFile, ABK_LOGFILE_ENTRY_DELEMITER, 
					(DWORD)_tcslen(ABK_LOGFILE_ENTRY_DELEMITER), &dwSavedSz, NULL);

				delete[] szPtr; // Free allocated memory in the cfListView_GetItemText

				szPtr = cfListView_GetItemText(g_hwndListView, i, 2);

				WriteFile(hFile, szPtr, (DWORD)_tcslen(szPtr), &dwSavedSz, NULL);

				delete[] szPtr; // Free allocated memory in the cfListView_GetItemText

				WriteFile(hFile, _T("\r\n"), (DWORD)_tcslen(_T("\r\n")), &dwSavedSz, NULL);
			};

			CloseHandle(hFile);
		}; // else of if(hFile == INVALID_HANDLE_VALUE)
	};
};

BOOL PutInClipboard(LPTSTR szStr_)
{
	BOOL fRet = FALSE;
	BOOL fOpen = FALSE;

	__try
	{
		LONG   nLen;
		LPVOID ptr;

		fOpen = OpenClipboard(g_hwndLogViewer);
		if(!fOpen) 
			__leave;

		if(!EmptyClipboard()) 
			__leave;

		nLen = (LONG)_tcslen(szStr_);

		// Put regular text block.
		g_hClipTextLV = GlobalAlloc(GMEM_MOVEABLE, (nLen + 1));
		if(!g_hClipTextLV)
			__leave;
		
		ptr = GlobalLock(g_hClipTextLV);
		_tcscpy((LPSTR)ptr, szStr_);
		GlobalUnlock(g_hClipTextLV);
		SetClipboardData(CF_TEXT, g_hClipTextLV);

		// Put OEM text block.
		g_hClipOEMLV = GlobalAlloc(GMEM_MOVEABLE, (nLen + 1));
		if(!g_hClipOEMLV)
			__leave;
		
		ptr = GlobalLock(g_hClipOEMLV);
		CharToOem(szStr_, (LPTSTR)ptr);
		GlobalUnlock(g_hClipOEMLV);
		SetClipboardData(CF_OEMTEXT, g_hClipOEMLV);

		// Put UNICODE text block.
		nLen = MultiByteToWideChar(CP_ACP, 0, szStr_, -1, NULL, 0);

		g_hClipUTFLV = GlobalAlloc(GMEM_MOVEABLE, (nLen + 1) * sizeof(WCHAR));
		if(!g_hClipOEMLV)
			__leave;

		ptr = GlobalLock(g_hClipUTFLV);
		MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, szStr_, -1, (LPWSTR)ptr, nLen);
		GlobalUnlock(g_hClipUTFLV);
		SetClipboardData(CF_UNICODETEXT, g_hClipUTFLV);

		fRet = TRUE;
	}
	__finally
	{
		if(fOpen)
			CloseClipboard();
	};

	return(fRet);
};

HWND CreateStatusbarLV(HWND hPrn_)
{
	g_hwndStatusbar = CreateStatusWindow(WS_CHILD|WS_VISIBLE, NULL, hPrn_, 0);

	return(g_hwndStatusbar);
};

HWND CreateToolbarLV(HWND hPrn_)
{
	DWORD dwExStyle;

	g_hwndToolbar = CreateWindowEx(WS_EX_TOOLWINDOW, TOOLBARCLASSNAME, (LPSTR)NULL, 
						WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS|WS_CLIPCHILDREN
						|TBSTYLE_FLAT|TBSTYLE_LIST|TBSTYLE_AUTOSIZE|TBSTYLE_TOOLTIPS
						|CCS_NODIVIDER|/*CCS_NOPARENTALIGN|CCS_NORESIZE|*/CCS_TOP, 
						0, 0, 0, 0, hPrn_, (HMENU)0, g_hInst, NULL);

	// Send the TB_BUTTONSTRUCTSIZE message, which is required for 
	// backward compatibility. 
	SendMessage(g_hwndToolbar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);

	// Modify extended toolbar styles.
	dwExStyle = (DWORD)SendMessage(g_hwndToolbar, TB_GETEXTENDEDSTYLE, 0, 0);
	SendMessage(g_hwndToolbar, TB_SETEXTENDEDSTYLE, 0, 
				  (LPARAM)dwExStyle|TBSTYLE_EX_MIXEDBUTTONS/*|TBSTYLE_EX_HIDECLIPPEDBUTTONS*/);

	// Add image lists.
	SendMessage(g_hwndToolbar, TB_SETIMAGELIST, 0, (LPARAM)g_hImgLstNorm);
	SendMessage(g_hwndToolbar, TB_SETHOTIMAGELIST, 0, (LPARAM)g_hImgLstHot);
	SendMessage(g_hwndToolbar, TB_SETDISABLEDIMAGELIST, 0, (LPARAM)g_hImgLstDis);

	return(g_hwndToolbar);
};

BOOL CreateImageListsLV()
{
	cfdprintf(_T("CreateImageListsLV"));

	// Create image lists.
	g_hImgLstTypes = ImageList_Create(16, 16, ILC_COLOR24|ILC_MASK, 4, 0);
	if(g_hImgLstTypes == NULL)
		return(FALSE);

	g_hImgLstNorm = ImageList_Create(ABKLV_TB_IMAGE_SIZE_X, ABKLV_TB_IMAGE_SIZE_Y, ILC_COLOR24|ILC_MASK, 8, 1);
	if(g_hImgLstNorm == NULL)
		return(FALSE);

	g_hImgLstHot  = ImageList_Create(ABKLV_TB_IMAGE_SIZE_X, ABKLV_TB_IMAGE_SIZE_Y, ILC_COLOR24|ILC_MASK, 8, 1);
	if(g_hImgLstHot == NULL)
		return(FALSE);

	g_hImgLstDis  = ImageList_Create(ABKLV_TB_IMAGE_SIZE_X, ABKLV_TB_IMAGE_SIZE_Y, ILC_COLOR8|ILC_MASK, 8, 1);
	if(g_hImgLstDis == NULL)
		return(FALSE);

	// Populate image lists with images.
	// Types images.
	if(!AddImageIntoImageListLV(g_hImgLstTypes, IDB_TYPE_INFO, IDB_TYPE_MASK))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstTypes, IDB_TYPE_OK, IDB_TYPE_MASK))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstTypes, IDB_TYPE_ERROR, IDB_TYPE_MASK))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstTypes, IDB_TYPE_MSG, IDB_TYPE_MASK))
		return(FALSE);

	// Add normal images.
	if(!AddImageIntoImageListLV(g_hImgLstNorm, IDB_SAVE_00, IDB_SAVE_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstNorm, IDB_OPEN_00, IDB_OPEN_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstNorm, IDB_DELETE_00, IDB_DELETE_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstNorm, IDB_COPY_00, IDB_COPY_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstNorm, IDB_SYNC_00, IDB_SYNC_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstNorm, IDB_FIND_00, IDB_FIND_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstNorm, IDB_FIND_NEXT_00, IDB_FIND_NEXT_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstNorm, IDB_AUTOSCROLL_00, IDB_AUTOSCROLL_MASK_00))
		return(FALSE);

	//// Add hot images.
	if(!AddImageIntoImageListLV(g_hImgLstHot, IDB_SAVE_HOT_00, IDB_SAVE_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstHot, IDB_OPEN_HOT_00, IDB_OPEN_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstHot, IDB_DELETE_HOT_00, IDB_DELETE_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstHot, IDB_COPY_HOT_00, IDB_COPY_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstHot, IDB_SYNC_HOT_00, IDB_SYNC_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstHot, IDB_FIND_HOT_00, IDB_FIND_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstHot, IDB_FIND_NEXT_HOT_00, IDB_FIND_NEXT_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstHot, IDB_AUTOSCROLL_HOT_00, IDB_AUTOSCROLL_MASK_00))
		return(FALSE);

	//// Add disabled images.
	if(!AddImageIntoImageListLV(g_hImgLstDis, IDB_SAVE_DIS_00, IDB_SAVE_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstDis, IDB_OPEN_DIS_00, IDB_OPEN_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstDis, IDB_DELETE_DIS_00, IDB_DELETE_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstDis, IDB_COPY_DIS_00, IDB_COPY_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstDis, IDB_SYNC_DIS_00, IDB_SYNC_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstDis, IDB_FIND_DIS_00, IDB_FIND_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstDis, IDB_FIND_NEXT_DIS_00, IDB_FIND_NEXT_MASK_00))
		return(FALSE);

	if(!AddImageIntoImageListLV(g_hImgLstDis, IDB_AUTOSCROLL_DIS_00, IDB_AUTOSCROLL_MASK_00))
		return(FALSE);

	cfdprintf(_T("!CreateImageListsLV"));

	return(TRUE);
};

BOOL AddImageIntoImageListLV(HIMAGELIST hImgLst_, DWORD dwImgID_, DWORD dwImgIDMask_)
{
	HBITMAP	hbm, hbms;
	int		nImg;

	cfdprintf(_T("AddImageIntoImageList"));

	hbm = (HBITMAP)LoadImage(g_hInst, MAKEINTRESOURCE(dwImgID_), 
							   IMAGE_BITMAP, 0, 0, LR_DEFAULTCOLOR);

	hbms = (HBITMAP)LoadImage(g_hInst, MAKEINTRESOURCE(dwImgIDMask_), 
							   IMAGE_BITMAP, 0, 0, LR_DEFAULTCOLOR);

	if(hbm == NULL || hbms == NULL)
	{
		cfdprintf(_T("AddImageIntoImageList: Failed to load image %d %d."), dwImgID_, GetLastError());
		return(FALSE);
	};

	nImg = ImageList_Add(hImgLst_, hbm, hbms);
	if(nImg == -1)
	{
		cfdprintf(_T("AddImageIntoImageList: Failed to add image %d %d."), dwImgID_, GetLastError());
		return(FALSE);
	};

	DeleteObject(hbm);
	DeleteObject(hbms);

	cfdprintf(_T("!AddImageIntoImageList"));

	return(TRUE);
};

BOOL AddToolbarButtonLV(HWND hwndTlb_, DWORD dwID_, DWORD dwImID_, DWORD dwData_, BYTE bStyle_, BYTE bState_)
{
	size_t			cch;
	CONST INT		cBufSize = 100;
	TCHAR			szBuf[cBufSize];
	LPTSTR			szPtr;
	TBBUTTON		tbb;
	TBBUTTONINFO	tbbi;

	if(dwID_ > 0)
	{
		// Load button label.
		if(!::LoadString(g_hInst, dwID_, szBuf, cBufSize-1))
			return -1;

		//Save room for second null terminator.
		cch = _tcslen(szBuf);
		szBuf[cch+2] = 0; //Double-null terminate.
		szPtr = szBuf;
	}
	else
	{
		szPtr = NULL;
	};

	// Add button.
	tbb.iBitmap   = MAKELONG(dwImID_, 0); //dwImID_;
	tbb.idCommand = dwID_;
	tbb.fsState   = bState_;
	tbb.fsStyle   = bStyle_;
	tbb.dwData    = dwData_;
	tbb.iString   = (INT_PTR)szPtr;

	::SendMessage(hwndTlb_, TB_ADDBUTTONS, 1, (LPARAM)&tbb);

	// Modify style.
	tbbi.cbSize = sizeof(TBBUTTONINFO);

	if(bStyle_ & BTNS_SEP)
	{
		tbbi.dwMask = TBIF_SIZE;
		tbbi.cx = 4;
		::SendMessage(hwndTlb_, TB_SETBUTTONINFO, dwID_, (LPARAM)&tbbi);
	};

	//::SendMessage(m_hwndToolbar, TB_AUTOSIZE, 0, 0);

	return(TRUE);
};

BOOL CreateRebarLV(HWND hPrn_)
{
	INITCOMMONCONTROLSEX icex;
	REBARBANDINFO rbBand;
	REBARINFO rbi;

	icex.dwSize = sizeof(INITCOMMONCONTROLSEX);
	icex.dwICC  = ICC_COOL_CLASSES|ICC_BAR_CLASSES;
	InitCommonControlsEx(&icex);

	g_hwndRebar = CreateWindowEx(WS_EX_TOOLWINDOW, REBARCLASSNAME, NULL,
								 WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS|WS_BORDER
								 |WS_CLIPCHILDREN|RBS_VARHEIGHT
								 |CCS_NODIVIDER|CCS_NOPARENTALIGN
								 |RBS_AUTOSIZE|RBS_BANDBORDERS,
								 0, 0, 0, 0,
								 hPrn_, NULL, g_hInst, NULL);

	if(!g_hwndRebar)
	{
		cfdprintf(_T("CreateRebarLV: Failed to create rebar window."));
		return(NULL);
	};

	rbi.cbSize = sizeof(REBARINFO);
	rbi.fMask  = 0;
	rbi.himl   = (HIMAGELIST)NULL;

	if(!SendMessage(g_hwndRebar, RB_SETBARINFO, 0, (LPARAM)&rbi))
		return(NULL);

	// Initialize structure members that both bands will share.
	rbBand.cbSize = sizeof(REBARBANDINFO);  // Required
	rbBand.fMask  = RBBIM_STYLE|RBBIM_CHILD|RBBIM_CHILDSIZE/*|RBBIM_SIZE*/;
	rbBand.fStyle = RBBS_BREAK/*|RBBS_CHILDEDGE*/|RBBS_NOGRIPPER;

	// Create and setup band for toolbar.
	if(!CreateToolbarLV(g_hwndRebar))
		return(FALSE);

	// Get the height of the toolbar.
	DWORD dwBtnSize = (DWORD)SendMessage(g_hwndToolbar, TB_GETBUTTONSIZE, 0,0);

	// Set values unique to the band with the combo box.
	rbBand.hwndChild  = g_hwndToolbar;
	rbBand.cxMinChild = 0;
	rbBand.cyMinChild = HIWORD(dwBtnSize);//+2;
	rbBand.cx         = 250;

	// Add the band that has the toolbar.
	SendMessage(g_hwndRebar, RB_INSERTBAND, (WPARAM)-1, (LPARAM)&rbBand);

	// Create and setup band for preparing progress bar.
	//m_hwndPB = ::CreateWindowEx(0, PROGRESS_CLASS, (LPSTR) NULL, 
	//						  WS_CHILD | WS_VISIBLE | PBS_SMOOTH,
	//						  0, 0, 0, 15, 
	//						  g_hwndRebar, (HMENU) 0, g_hInst, NULL);

	//// Set values unique to the band with the combo box.
	//rbBand.fMask  |= RBBIM_TEXT;
	//rbBand.fStyle |= RBBS_CHILDEDGE;
	//rbBand.lpText     = _T("Preparing");
	//rbBand.hwndChild  = m_hwndPB;
	//rbBand.cxMinChild = 0;
	//rbBand.cyMinChild = 15;
	//rbBand.cx         = 200;

	//// Add the band that has the toolbar.
	//::SendMessage(m_hwndRB, RB_INSERTBAND, (WPARAM)-1, (LPARAM)&rbBand);

	//this->ShowProgressBand(FALSE);

	return TRUE;
};

VOID ShowToolbarLV(BOOL fShow_)
{
	RECT rc;

	SendMessage(g_hwndRebar, RB_SHOWBAND, 0, fShow_);

	// Hack ;-)
	OnSizeLV(g_hwndLogViewer, 0, 0);
	GetClientRect(g_hwndListView, &rc);
	InvalidateRect(g_hwndListView, &rc, TRUE);
};

VOID ShowStatusbarLV(BOOL fShow_)
{
	ShowWindow(g_hwndStatusbar, fShow_ ? SW_SHOW : SW_HIDE);
	OnSizeLV(g_hwndLogViewer, 0, 0); // Hack ;-)
};

HWND CreateListViewLV(HWND hPrn_)
{
	DWORD		dwExStyle;
	LVCOLUMN	lvc;

	g_hwndListView = CreateWindow(WC_LISTVIEW, NULL, 
		WS_CHILD|WS_VISIBLE|LVS_REPORT|LVS_NOSORTHEADER
		|LVS_SHOWSELALWAYS/*|LVS_SINGLESEL*//*|LVS_OWNERDRAWFIXED*/, 
		0, 0, 0, 0, hPrn_, NULL, g_hInst, NULL);

	if(!g_hwndListView)
	{
		cfdprintf(_T("CreateListViewLV: Cteate ListView failed."));
		return(NULL);
	};

	dwExStyle = GetWindowLong(g_hwndListView, GWL_EXSTYLE);

	SetWindowLong(g_hwndListView, GWL_EXSTYLE, dwExStyle|WS_EX_CLIENTEDGE);

	ListView_SetExtendedListViewStyle(g_hwndListView, LVS_EX_FULLROWSELECT|LVS_EX_LABELTIP);

	// Add image list.
	ListView_SetImageList(g_hwndListView, g_hImgLstTypes, LVSIL_SMALL);

	// Add columns.
	lvc.mask = LVCF_SUBITEM|LVCF_TEXT|LVCF_WIDTH;
	lvc.cx = 25;
	lvc.pszText = _T("");
	lvc.iSubItem = 0;

	SendMessage(g_hwndListView, LVM_INSERTCOLUMN, 0, (LPARAM)&lvc);

	lvc.cx = g_dwFirstColWidth;
	lvc.pszText = _T("/");
	lvc.iSubItem = 1;

	SendMessage(g_hwndListView, LVM_INSERTCOLUMN, 1, (LPARAM)&lvc);

	lvc.cx = g_dwSecondColWidth;
	lvc.pszText = _T(" ");
	lvc.iSubItem = 2;

	SendMessage(g_hwndListView, LVM_INSERTCOLUMN, 2, (LPARAM)&lvc);

	return(g_hwndListView);
};

VOID AddListViewLineLV(DWORD dwType_, LPSYSTEMTIME lpst_, LPTSTR szText_)
{
	TCHAR	szTimeDateBuf[20];

	_stprintf(szTimeDateBuf, _T("%02d-%02d-%04d %02d:%02d:%02d"), 
				lpst_->wDay, lpst_->wMonth, lpst_->wYear, 
				lpst_->wHour, lpst_->wMinute, lpst_->wSecond);

	AddListViewLineLV(dwType_, szTimeDateBuf, szText_);
};

VOID AddListViewLineLV(DWORD dwType_, LPTSTR szTime_, LPTSTR szText_)
{
	LVITEM	lvi;
	DWORD	dwItmCnt;

	dwItmCnt = (DWORD)SendMessage(g_hwndListView, LVM_GETITEMCOUNT, 0, 0);

	// Add image.
	lvi.mask = LVIF_PARAM|LVIF_STATE|LVIF_IMAGE;
	lvi.state = 0;
	lvi.stateMask = 0;
	lvi.iItem = dwItmCnt;
	lvi.iImage = dwType_ - 1;

	lvi.iSubItem = 0;
	lvi.lParam = dwType_;

	// Add subitems text.
	lvi.iItem = (INT)SendMessage(g_hwndListView, LVM_INSERTITEM, 0, (LPARAM)&lvi);

	lvi.mask = LVIF_TEXT;

	lvi.iSubItem = 1;
	lvi.pszText = szTime_;

	SendMessage(g_hwndListView, LVM_SETITEM, 0, (LPARAM)&lvi);

	lvi.iSubItem = 2;
	lvi.pszText = szText_;

	SendMessage(g_hwndListView, LVM_SETITEM, 0, (LPARAM)&lvi);

	if((dwItmCnt + 1) > g_dwMaxEventsShow && g_dwMaxEventsShow > 0 && g_fSvcSync)
		SendMessage(g_hwndListView, LVM_DELETEITEM, 0, 0);
	else 
		dwItmCnt++;

	if(g_fAutoscroll)
	{
		cfdprintf(_T("Autoscroll: %d; %d"), dwItmCnt, ListView_EnsureVisible(g_hwndListView, dwItmCnt-1, FALSE));
	};
};


