/*  Botik::Key Pro - Botik Key Client for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2004 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "opt_tab_access_dlg.h"
#include "main_dlg.h"
#include "message_dispatcher.h"

HWND g_hAccTabDlg;

// Forward declaration.
VOID SetAccessModeDescr(HWND hWnd_, DWORD dwMode_);

HWND CreateAccTab(HWND hPrnWnd_)
{
	g_hAccTabDlg = CreateDialog(g_hInst, MAKEINTRESOURCE(IDD_OPT_TAB_ACCESS), 
								hPrnWnd_, (DLGPROC)WndProcAT);
	return(g_hAccTabDlg);
};

LRESULT CALLBACK WndProcAT(HWND hWnd_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_)
{
	PAINTSTRUCT ps;
	HDC hdc;
	DWORD dwPos, dwAcc = 0;

	switch(uMsg_) 
	{
	case WM_INITDIALOG:
		return(OnInitDialogAT(hWnd_, wParam_, lParam_));

	case WM_COMMAND:
		return(OnCommandAT(hWnd_, wParam_, lParam_));

	case WM_VSCROLL:
		dwPos = (DWORD)SendDlgItemMessage(hWnd_, IDC_ACCESS_SLIDER, TBM_GETPOS, 0, 0); 

		switch(LOWORD(wParam_)) 
		{
		case TB_THUMBTRACK:
			cfdprintf("Track. Pos: %d; Mode: %d", dwPos, (1 << (dwPos - 1)));
			SetAccessModeDescr(hWnd_, (1 << (dwPos - 1)));
			break;

		case TB_ENDTRACK:
			cfdprintf("End. Pos: %d; Mode: %d", dwPos, (1 << (dwPos - 1)));
			SetAccessModeDescr(hWnd_, (1 << (dwPos - 1)));
			SetAccessMode((1 << (dwPos - 1)));
			break;
		};
		break; // case WM_VSCROLL

	case WM_PAINT:
		hdc = BeginPaint(hWnd_, &ps);
		EndPaint(hWnd_, &ps);
		break;
	};

	return(FALSE);
};

VOID SetAccessModeDescr(HWND hWnd_, DWORD dwMode_)
{
	TCHAR szText[MAX_LOADSTRING];
	UINT  uiIDN, uiIDD = 0;

	if(dwMode_ & ABK_ACCESS_NONE)
	{
		uiIDN = IDS_MODE_NAME_01;
		uiIDD = IDS_MODE_DESCR_01;
	}
	else if(dwMode_ & ABK_ACCESS_LOCAL)
	{
		uiIDN = IDS_MODE_NAME_02;
		uiIDD = IDS_MODE_DESCR_02;
	}
	else if(dwMode_ & ABK_ACCESS_WORLD)
	{
		uiIDN = IDS_MODE_NAME_03;
		uiIDD = IDS_MODE_DESCR_03;
	}
	else
	{
		uiIDN = IDS_MODE_NAME_00;
		uiIDD = IDS_MODE_DESCR_00;
	};

	LoadString(g_hInst, uiIDN, szText, MAX_LOADSTRING-1);
	SetDlgItemText(hWnd_, IDC_MODE_NAME, szText);

	LoadString(g_hInst, uiIDD, szText, MAX_LOADSTRING-1);
	SetDlgItemText(hWnd_, IDC_MODE_DESCR, szText);
};

LRESULT OnInitDialogAT(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	LOGFONT lf;
	HICON hIcon;

	// Get default GUI font information.
	GetObject(GetStockObject(DEFAULT_GUI_FONT), sizeof(LOGFONT), &lf);

	// Change some params.
	lf.lfWeight = FW_BOLD;

	HFONT hFont = CreateFont(lf.lfHeight, lf.lfWidth, lf.lfEscapement, 
							 lf.lfOrientation, lf.lfWeight, lf.lfItalic, 
							 lf.lfUnderline, lf.lfStrikeOut, lf.lfCharSet, 
							 lf.lfOutPrecision, lf.lfClipPrecision, lf.lfQuality, 
							 lf.lfPitchAndFamily, lf.lfFaceName);

	SendDlgItemMessage(hWnd_, IDC_MODE_NAME, WM_SETFONT, (WPARAM)hFont, TRUE);

	SendDlgItemMessage(hWnd_, IDC_ACCESS_SLIDER, TBM_SETLINESIZE, 0, 1);
	SendDlgItemMessage(hWnd_, IDC_ACCESS_SLIDER, TBM_SETRANGE, TRUE, MAKELONG(1, 3));
	SendDlgItemMessage(hWnd_, IDC_ACCESS_SLIDER, TBM_SETTIC, 0, 2);

	hIcon = (HICON)LoadImage(g_hInst, MAKEINTRESOURCE(IDI_SVC_START), IMAGE_ICON, 16, 16, 0);
	SendDlgItemMessage(hWnd_, IDC_SVC_START, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);

	hIcon = (HICON)LoadImage(g_hInst, MAKEINTRESOURCE(IDI_SVC_STOP), IMAGE_ICON, 16, 16, 0);
	SendDlgItemMessage(hWnd_, IDC_SVC_STOP, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);

	cfdprintf("Before DisplayCurrentAccessMode");
	DisplayCurrentAccessMode(hWnd_);

	DisplayCurrentServiceStatus(hWnd_);

	return TRUE;
};

LRESULT OnCommandAT(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	cfdprintf("WM_COMMAND");

	switch(LOWORD(wParam_))
	{
	case IDC_SVC_START:
		if(HIWORD(wParam_) == BN_CLICKED)
		{
			DoStartupService();
		};
		break;
	case IDC_SVC_STOP:
		if(HIWORD(wParam_) == BN_CLICKED)
		{
			DoShutdownService();
		};
		break;
	};
	return(0);
};

VOID DisplayCurrentServiceStatus(HWND hWnd_)
{
	UINT nID = 0;
	TCHAR szText[MAX_LOADSTRING];

	cfdprintf("DisplayCurrentServiceStatus: 0x%x", g_dwSvcStatus);

	if((g_dwSvcStatus & ABK_SVC_STATUS_MASK) == ABK_SVC_STOPPED)
	{
		cfEnableDlgItem(hWnd_, IDC_SVC_START, TRUE);
		cfEnableDlgItem(hWnd_, IDC_SVC_STOP, FALSE);
		nID = IDS_STAT_DESCR_0000;
	}
	else if((g_dwSvcStatus & ABK_SVC_STATUS_MASK) == ABK_SVC_RUNNING)
	{
		cfEnableDlgItem(hWnd_, IDC_SVC_STOP, TRUE);
		cfEnableDlgItem(hWnd_, IDC_SVC_START, FALSE);
		nID = IDS_STAT_DESCR_0001;
	}
	else
	{
		cfEnableDlgItem(hWnd_, IDC_SVC_STOP, FALSE);
		cfEnableDlgItem(hWnd_, IDC_SVC_START, FALSE);
		nID = IDS_STAT_DESCR_0002;
	};

	if((g_dwSvcStatus & ABK_SVC_BUSINESS_MASK) == ABK_SVC_CONNECTING)
		nID = IDS_STAT_DESCR_0003;
	else if((g_dwSvcStatus & ABK_SVC_BUSINESS_MASK) == ABK_SVC_CONNECTED)
		nID = IDS_STAT_DESCR_0004;

	LoadString(g_hInst, nID, szText, MAX_LOADSTRING-1);
	SetDlgItemText(hWnd_, IDC_STATE_DESCR, szText);
};

VOID DisplayCurrentAccessMode(HWND hWnd_)
{
	INT nIconID;
	UINT uiImgID;
	HANDLE hImg;

    cfdprintf("DisplayCurrentAccessMode: g_nCurAccess: %d; g_nReqAccess: %d", g_nCurAccess, g_nReqAccess);

    // we need to show current access mode in the traffic light
    // and in the tray icon, but we must set slider to the 
    // the requested position. i.e. if current access mode was setted
    // as server to none, but we requested world, we show none mode
    // in the traffic light and tray icon, but slider is in the world 
    // position

	// Current access mode.
	if(g_nReqAccess & ABK_ACCESS_NONE)
	{
		cfEnableDlgItem(hWnd_, IDC_ACCESS_SLIDER, TRUE);
		ShowWindow(GetDlgItem(g_hAccTabDlg, IDC_ACCESS_SLIDER), SW_SHOW);
		ShowWindow(GetDlgItem(g_hAccTabDlg, IDC_TEXT_NOTE), SW_SHOW);
		SendDlgItemMessage(hWnd_, IDC_ACCESS_SLIDER, TBM_SETPOS, TRUE, 1);
	}
	else if(g_nReqAccess & ABK_ACCESS_LOCAL)
	{
		cfEnableDlgItem(hWnd_, IDC_ACCESS_SLIDER, TRUE);
		ShowWindow(GetDlgItem(g_hAccTabDlg, IDC_ACCESS_SLIDER), SW_SHOW);
		ShowWindow(GetDlgItem(g_hAccTabDlg, IDC_TEXT_NOTE), SW_SHOW);
		SendDlgItemMessage(hWnd_, IDC_ACCESS_SLIDER, TBM_SETPOS, TRUE, 2);
	}
	else if(g_nReqAccess & ABK_ACCESS_WORLD)
	{
		cfEnableDlgItem(hWnd_, IDC_ACCESS_SLIDER, TRUE);
		ShowWindow(GetDlgItem(g_hAccTabDlg, IDC_ACCESS_SLIDER), SW_SHOW);
		ShowWindow(GetDlgItem(g_hAccTabDlg, IDC_TEXT_NOTE), SW_SHOW);
		SendDlgItemMessage(hWnd_, IDC_ACCESS_SLIDER, TBM_SETPOS, TRUE, 3);
	}
	else
	{
		ShowWindow(GetDlgItem(g_hAccTabDlg, IDC_ACCESS_SLIDER), SW_HIDE);
		ShowWindow(GetDlgItem(g_hAccTabDlg, IDC_TEXT_NOTE), SW_HIDE);
	};

    if(g_nCurAccess & ABK_ACCESS_NONE)
	{
		nIconID = IDI_NET_NONE;
		uiImgID = IDB_TRLIGHT_RED;
	}
	else if(g_nCurAccess & ABK_ACCESS_LOCAL)
	{
		nIconID = IDI_NET_LOCAL;
		uiImgID = IDB_TRLIGHT_GREEN_L;
	}
	else if(g_nCurAccess & ABK_ACCESS_WORLD)
	{
		nIconID = IDI_NET_WORLD;
		uiImgID = IDB_TRLIGHT_GREEN_W;
	}
	else
	{
		nIconID = IDI_MANAGER;
		uiImgID = IDB_TRLIGHT_OFF;
	};

	SetAccessModeDescr(hWnd_, g_nReqAccess);

	// If we are in a changing process to the mode g_nCurAccess,
	// show yellow light and botik icon.
	if((g_nCurAccess != ABK_ACCESS_UNSPEC) && (g_nCurAccess & ABK_ACCESS_CHANGING))
	{
		cfEnableDlgItem(g_hAccTabDlg, IDC_ACCESS_SLIDER, FALSE);
		uiImgID = IDB_TRLIGHT_YELLOW;
		nIconID = IDI_NET_YELLOW;
	};

	hImg = LoadImage(g_hInst, MAKEINTRESOURCE(uiImgID), IMAGE_BITMAP, 0, 0, LR_DEFAULTSIZE);
	hImg = (HANDLE)SendDlgItemMessage(hWnd_, IDC_SEMAPHORE, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hImg);
	if(hImg)
		DeleteObject(hImg);

	ModNotifyIcon(g_hMainDlg, ABK_MNGR_NOTIFY_ICON_ID, nIconID);
};


